import asyncio
import os
from typing import Callable, List, Optional, Tuple, Union

from tqdm import tqdm

from supervisely import get_project_class
from supervisely._utils import run_coroutine
from supervisely.annotation.annotation import Annotation, ProjectMeta
from supervisely.api.api import Api
from supervisely.api.dataset_api import DatasetInfo
from supervisely.api.image_api import ImageInfo
from supervisely.api.project_api import ProjectInfo
from supervisely.io.env import apps_cache_dir
from supervisely.io.fs import (
    copy_dir_recursively,
    copy_file,
    dir_exists,
    get_directory_size,
    remove_dir,
)
from supervisely.io.json import load_json_file
from supervisely.project import Project
from supervisely.project.project import Dataset, OpenMode, ProjectType
from supervisely.sly_logger import logger


def download(
    api: Api,
    project_id: int,
    dest_dir: str,
    dataset_ids: Optional[List[int]] = None,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
    **kwargs,
) -> None:
    """
    Downloads project of any type to the local directory. See methods `sly.download_project`,
    `sly.download_video_project`, `sly.download_volume_project`, `sly.download_pointcloud_project`,
    `sly.download_pointcloud_episode_project` to examine full list of possible arguments.

    :param api: Supervisely API address and token.
    :type api: :class:`~supervisely.api.api.Api`
    :param project_id: Project ID, which will be downloaded.
    :type project_id: int
    :param dest_dir: Destination path to local directory.
    :type dest_dir: str
    :param dataset_ids: Specified list of Dataset IDs which will be downloaded.
    :type dataset_ids: list(int), optional
    :param log_progress: Show downloading logs in the output.
    :type log_progress: bool
    :param progress_cb: Function for tracking download progress.
    :type progress_cb: tqdm or callable, optional

    :returns: None.
    :rtype: NoneType

    :Usage Example:

        .. code-block:: python

            import os
            from tqdm import tqdm
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
                load_dotenv(os.path.expanduser("~/supervisely.env"))

            api = sly.Api.from_env()

            dest_dir = 'your/local/dest/dir'

            # Download image project
            project_id_image = 17732
            project_info = api.project.get_info_by_id(project_id_image)
            num_images = project_info.items_count

            p = tqdm(desc="Downloading image project", total=num_images)
            sly.download(
                api,
                project_id_image,
                dest_dir,
                progress_cb=p,
                save_image_info=True,
                save_images=True,
            )

            # Download video project
            project_id_video = 60498
            project_info = api.project.get_info_by_id(project_id_video)
            num_videos = project_info.items_count

            # Download video project with automatic logging...
            sly.download(
                api,
                project_id_video,
                dest_dir,
                save_video_info=True,
            )

            # ...or disable logging at all
            sly.download(
                api,
                project_id_video,
                dest_dir,
                log_progress=False,
                save_video_info=True,
            )

            # Download volume project
            project_id_volume = 18594
            project_info = api.project.get_info_by_id(project_id_volume)
            num_volumes = project_info.items_count

            p = tqdm(desc="Downloading volume project",total=num_volumes)
            sly.download(
                api,
                project_id_volume,
                dest_dir,
                progress_cb=p,
                download_volumes=True,
            )

            # Download pointcloud project
            project_id_ptcl = 18592
            project_info = api.project.get_info_by_id(project_id_ptcl)
            num_ptcl = project_info.items_count

            p = tqdm(desc="Downloading pointcloud project", total=num_ptcl)
            sly.download(
                api,
                project_id_ptcl,
                dest_dir,
                progress_cb=p,
                download_pointclouds_info=True,
            )

            # Download some datasets from pointcloud episodes project
            project_id_ptcl_ep = 18593
            dataset_ids = [43546, 45765, 45656]

            p = tqdm(
                desc="Download some datasets from pointcloud episodes project",
                total=len(dataset_ids),
            )
            sly.download(
                api,
                project_id_ptcl_ep,
                dest_dir,
                dataset_ids,
                progress_cb=p,
                download_pcd=True,
                download_related_images=True,
                download_annotations=True,
                download_pointclouds_info=True,
            )
    """

    project_info = api.project.get_info_by_id(project_id)

    if progress_cb is not None:
        log_progress = False

    project_class = get_project_class(project_info.type)
    project_class.download(
        api=api,
        project_id=project_id,
        dest_dir=dest_dir,
        dataset_ids=dataset_ids,
        log_progress=log_progress,
        progress_cb=progress_cb,
        **kwargs,
    )


def download_async(
    api: Api,
    project_id: int,
    dest_dir: str,
    semaphore: Optional[asyncio.Semaphore] = None,
    dataset_ids: Optional[List[int]] = None,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
    **kwargs,
) -> None:
    project_info = api.project.get_info_by_id(project_id)

    if progress_cb is not None:
        log_progress = False

    project_class = get_project_class(project_info.type)
    if hasattr(project_class, "download_async"):
        download_coro = project_class.download_async(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            semaphore=semaphore,
            dataset_ids=dataset_ids,
            log_progress=log_progress,
            progress_cb=progress_cb,
            **kwargs,
        )
        run_coroutine(download_coro)
    else:
        raise NotImplementedError(f"Method download_async is not implemented for {project_class}")


def download_async_or_sync(
    api: Api,
    project_id: int,
    dest_dir: str,
    dataset_ids: Optional[List[int]] = None,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
    semaphore: Optional[Union[asyncio.Semaphore, int]] = None,
    **kwargs,
):
    """
    Download project asynchronously if possible, otherwise download synchronously.
    Automatically detects project type.
    You can pass project info object as `project_info` kwarg to avoid additional API requests.

    In case of error during asynchronous download, the function will switch to synchronous download.

    :param api: Supervisely API address and token.
    :type api: :class:`~supervisely.api.api.Api`
    :param project_id: Project ID.
    :type project_id: int
    :param dest_dir: Destination directory.
    :type dest_dir: str
    :param dataset_ids: Dataset IDs.
    :type dataset_ids: List[int], optional
    :param log_progress: Show progress bar.
    :type log_progress: bool
    :param progress_cb: Progress callback.
    :type progress_cb: tqdm or callable, optional
    :param semaphore: Semaphore to limit the number of concurrent downloads.
    :type semaphore: asyncio.Semaphore, optional
    :param kwargs: Additional keyword arguments.
    :type kwargs: dict
    """
    project_info = kwargs.pop("project_info", None)
    if not isinstance(project_info, ProjectInfo) or project_info.id != project_id:
        project_info = api.project.get_info_by_id(project_id)

    if progress_cb is not None:
        log_progress = False

    project_class = get_project_class(project_info.type)

    switch_to_sync = False
    if hasattr(project_class, "download_async"):
        try:
            download_coro = project_class.download_async(
                api=api,
                project_id=project_id,
                dest_dir=dest_dir,
                semaphore=semaphore,
                dataset_ids=dataset_ids,
                log_progress=log_progress,
                progress_cb=progress_cb,
                **kwargs,
            )
            run_coroutine(download_coro)
        except Exception as e:
            if kwargs.get("resume_download", False) is False:
                remove_dir(dest_dir)
            logger.error(f"Failed to download project {project_id} asynchronously: {e}")
            logger.warning("Switching to synchronous download")
            switch_to_sync = True
    else:
        switch_to_sync = True

    if switch_to_sync:
        project_class.download(
            api=api,
            project_id=project_id,
            dest_dir=dest_dir,
            dataset_ids=dataset_ids,
            log_progress=log_progress,
            progress_cb=progress_cb,
            **kwargs,
        )


def download_fast(
    api: Api,
    project_id: int,
    dest_dir: str,
    dataset_ids: Optional[List[int]] = None,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
    semaphore: Optional[asyncio.Semaphore] = None,
    **kwargs,
) -> None:
    """
    Download project in a fast mode.
    Items are downloaded asynchronously. If an error occurs, the method will fallback to synchronous download.
    Automatically detects project type.

    :param api: Supervisely API address and token.
    :type api: :class:`~supervisely.api.api.Api`
    :param project_id: Supervisely downloadable project ID.
    :type project_id: int
    :param dest_dir: Destination directory.
    :type dest_dir: str
    :param dataset_ids: Filter datasets by IDs.
    :type dataset_ids: List[int], optional
    :param log_progress: Show uploading progress bar.
    :type log_progress: bool
    :param progress_cb: Function for tracking download progress.
    :type progress_cb: tqdm or callable, optional
    :param semaphore: Semaphore to limit the number of concurrent downloads of items.
    :type semaphore: asyncio.Semaphore, optional
    :param only_image_tags: Download project with only images tags (without objects tags).
    :type only_image_tags: bool, optional
    :param save_image_info: Download images infos or not.
    :type save_image_info: bool, optional
    :param save_images: Download images or not.
    :type save_images: bool, optional
    :param save_image_meta: Download images metadata in JSON format or not.
    :type save_image_meta: bool, optional
    :param images_ids: Filter images by IDs.
    :type images_ids: List[int], optional
    :param resume_download: Resume download enables to download only missing files avoiding erase of existing files.
    :type resume_download: bool, optional
    :param switch_size: Size threshold that determines how an item will be downloaded.
                        Items larger than this size will be downloaded as single files, while smaller items will be downloaded as a batch.
                        Useful for projects with different item sizes and when you exactly know which size will perform better with batch download.
    :type switch_size: int, optional
    :param batch_size: Number of items to download in a single batch.
    :type batch_size: int, optional
    :param download_blob_files: Download project with Blob files in native format.
                                If False - download project like a regular project in classic Supervisely format.
    :type download_blob_files: bool, optional
    :param project_info: Project info object. To avoid additional API requests.
    :type project_info: :class:`~supervisely.api.project_api.ProjectInfo`, optional
    :param skip_create_readme: Skip creating README.md file. Default is False.
    :type skip_create_readme: bool, optional
    :returns: None
    :rtype: NoneType

    :Usage Example:

        .. code-block:: python

            import os
            from dotenv import load_dotenv

            import supervisely as sly

            # Load secrets and create API object from .env file (recommended)
            # Learn more here: https://developer.supervisely.com/getting-started/basics-of-authentication
            if sly.is_development():
                load_dotenv(os.path.expanduser("~/supervisely.env"))

            api = sly.Api.from_env()

            project_id = 8888
            save_directory = "/path/to/save/projects"

            sly.download_fast(api, project_id, save_directory)
    """

    download_async_or_sync(
        api=api,
        project_id=project_id,
        dest_dir=dest_dir,
        dataset_ids=dataset_ids,
        log_progress=log_progress,
        progress_cb=progress_cb,
        semaphore=semaphore,
        **kwargs,
    )


def _get_cache_dir(project_id: int, dataset_path: str = None) -> str:
    p = os.path.join(apps_cache_dir(), str(project_id))
    if dataset_path is not None:
        p = os.path.join(p, dataset_path)
    return p


def is_cached(project_id, dataset_path: str = None) -> bool:
    return dir_exists(_get_cache_dir(project_id, dataset_path))


def _split_by_cache(project_id: int, dataset_paths: List[str]) -> Tuple[List, List]:
    if not is_cached(project_id):
        return dataset_paths, []
    to_download = [ds_path for ds_path in dataset_paths if not is_cached(project_id, ds_path)]
    cached = [ds_path for ds_path in dataset_paths if is_cached(project_id, ds_path)]
    return to_download, cached


def get_cache_size(project_id: int, dataset_path: str = None) -> int:
    if not is_cached(project_id, dataset_path):
        return 0
    cache_dir = _get_cache_dir(project_id, dataset_path)
    return get_directory_size(cache_dir)


def _get_items_infos(api: Api, project_type: str, dataset_id: int) -> List[ImageInfo]:
    funcs = {
        str(ProjectType.IMAGES): api.image.get_list,
        str(ProjectType.VIDEOS): api.video.get_list,
        str(ProjectType.POINT_CLOUDS): api.pointcloud.get_list,
        str(ProjectType.POINT_CLOUD_EPISODES): api.pointcloud.get_list,
        str(ProjectType.VOLUMES): api.volume.get_list,
    }
    return funcs[project_type](dataset_id)


def _project_meta_changed(meta1: ProjectMeta, meta2: ProjectMeta) -> bool:
    if len(meta1.obj_classes) != len(meta2.obj_classes) or len(meta1.tag_metas) != len(
        meta2.tag_metas
    ):
        return True
    for obj_class1 in meta1.obj_classes:
        obj_class2 = meta2.get_obj_class(obj_class1.name)
        if obj_class2 is None or obj_class1 != obj_class2 or obj_class1.sly_id != obj_class2.sly_id:
            return True
    for tag_meta1 in meta1.tag_metas:
        tag_meta2 = meta2.get_tag_meta(tag_meta1.name)
        if tag_meta2 is None or tag_meta1 != tag_meta2 or tag_meta1.sly_id != tag_meta2.sly_id:
            return True
    return False


def _get_ds_full_name(
    dataset_info: DatasetInfo, all_ds_infos: List[DatasetInfo], suffix: str = ""
) -> str:
    if dataset_info.parent_id is None:
        return dataset_info.name + suffix
    parent = next((ds_info for ds_info in all_ds_infos if ds_info.id == dataset_info.parent_id))
    return _get_ds_full_name(parent, all_ds_infos, "/" + dataset_info.name + suffix)


def _validate_dataset(
    api: Api,
    project_id: int,
    project_type: str,
    project_meta: ProjectMeta,
    dataset_info: DatasetInfo,
    all_ds_infos: List[DatasetInfo] = None,
):
    try:
        project_class = get_project_class(project_type)
        project: Project = project_class(_get_cache_dir(project_id), OpenMode.READ)
    except Exception:
        logger.debug("Validating dataset failed. Error reading project.", exc_info=True)
        return False
    try:
        items_infos_dict = {
            item_info.name: item_info
            for item_info in _get_items_infos(api, project_type, dataset_info.id)
        }
    except:
        logger.debug("Validating dataset failed. Unable to download items infos.", exc_info=True)
        return False
    if all_ds_infos is None:
        all_ds_infos = api.dataset.get_list(project_id, recursive=True)
    project_meta_changed = _project_meta_changed(project_meta, project.meta)
    for dataset in project.datasets:
        dataset: Dataset
        if dataset.name == _get_ds_full_name(dataset_info, all_ds_infos):
            diff = set(items_infos_dict.keys()).difference(set(dataset.get_items_names()))
            if diff:
                logger.debug(
                    "Validating dataset failed. Items are missing.",
                    extra={"missing_items": diff},
                )
                return False
            for item_name, _, ann_path in dataset.items():
                try:
                    item_info = dataset.get_item_info(item_name)
                except Exception:
                    logger.debug(
                        "Validating dataset failed. Error reading item info.",
                        extra={"item_name": item_name},
                        exc_info=True,
                    )
                    return False
                if item_info.name not in items_infos_dict:
                    logger.debug(
                        "Validating dataset failed. Item info is redundant.",
                        extra={"item_name": item_name},
                    )
                    return False
                if item_info != items_infos_dict[item_info.name]:
                    logger.debug(
                        "Validating dataset failed. Item info is different.",
                        extra={"item_name": item_name},
                    )
                    return False
                if project_meta_changed:
                    try:
                        Annotation.from_json(load_json_file(ann_path), project_meta)
                    except Exception:
                        logger.debug(
                            "Validating dataset failed. Error reading annotation.",
                            extra={"item_name": item_name},
                            exc_info=True,
                        )
                        return False
            return True
    logger.debug(
        "Validating dataset failed. Dataset is missing.", extra={"dataset_name": dataset_info.name}
    )
    return False


def _validate(
    api: Api,
    project_info: ProjectInfo,
    project_meta: ProjectMeta,
    dataset_infos: List[DatasetInfo],
    all_ds_infos: List[DatasetInfo] = None,
):
    project_id = project_info.id
    to_download, cached = _split_by_cache(
        project_id, [get_dataset_path(api, dataset_infos, info.id) for info in dataset_infos]
    )
    to_download, cached = set(to_download), set(cached)
    for dataset_info in dataset_infos:
        ds_path = get_dataset_path(api, dataset_infos, dataset_info.id)
        if ds_path in to_download:
            continue
        if not _validate_dataset(
            api,
            project_id,
            project_info.type,
            project_meta,
            dataset_info,
            all_ds_infos,
        ):
            to_download.add(ds_path)
            cached.remove(ds_path)
            logger.info(
                f"Dataset {ds_path} of project {project_id} is not up to date and will be re-downloaded."
            )
    return list(to_download), list(cached)


def _add_save_items_infos_to_kwargs(kwargs: dict, project_type: str):
    arg_name = {
        str(ProjectType.IMAGES): "save_image_info",
        str(ProjectType.VIDEOS): "save_video_info",
        str(ProjectType.POINT_CLOUDS): "download_pointclouds_info",
        str(ProjectType.POINT_CLOUD_EPISODES): "download_pointclouds_info",
        str(ProjectType.VOLUMES): "save_volumes_info",
    }
    kwargs[arg_name[project_type]] = True
    return kwargs


def _add_resume_download_to_kwargs(kwargs: dict, project_type: str):
    supported_force_projects = (str(ProjectType.IMAGES), (str(ProjectType.VIDEOS)))
    if project_type in supported_force_projects:
        kwargs["resume_download"] = True
    return kwargs


def _download_project_to_cache(
    api: Api,
    project_info: ProjectInfo,
    dataset_infos: List[DatasetInfo],
    log_progress: bool = True,
    progress_cb: Callable = None,
    semaphore: Optional[asyncio.Semaphore] = None,
    **kwargs,
):
    project_id = project_info.id
    project_type = project_info.type
    kwargs = _add_save_items_infos_to_kwargs(kwargs, project_type)
    kwargs = _add_resume_download_to_kwargs(kwargs, project_type)
    cached_project_dir = _get_cache_dir(project_id)
    if len(dataset_infos) == 0:
        logger.debug("No datasets to download")
        return
    download_fast(
        api=api,
        project_id=project_id,
        dest_dir=cached_project_dir,
        dataset_ids=[info.id for info in dataset_infos],
        log_progress=log_progress,
        progress_cb=progress_cb,
        semaphore=semaphore,
        **kwargs,
    )


def download_to_cache(
    api: Api,
    project_id: int,
    dataset_infos: List[DatasetInfo] = None,
    dataset_ids: List[int] = None,
    log_progress: bool = True,
    progress_cb=None,
    semaphore: Optional[asyncio.Semaphore] = None,
    **kwargs,
) -> Tuple[List, List]:
    """
    Download datasets to cache.
    If dataset_infos is not None, dataset_ids must be None and vice versa.
    If both dataset_infos and dataset_ids are None, all datasets of the project will be downloaded.

    :param api: Supervisely API address and token.
    :type api: :class:`~supervisely.api.api.Api`
    :param project_id: Project ID, which will be downloaded.
    :type project_id: int
    :param dataset_infos: Specified list of Dataset Infos which will be downloaded.
    :type dataset_infos: List[:class:`~supervisely.api.dataset_api.DatasetInfo`], optional
    :param dataset_ids: Specified list of Dataset IDs which will be downloaded.
    :type dataset_ids: List[int], optional
    :param log_progress: Show downloading logs in the output.
    :type log_progress: bool, optional
    :param progress_cb: Function for tracking download progress. Will be called with number of items downloaded.
    :type progress_cb: tqdm or callable, optional
    :param semaphore: Semaphore for limiting the number of concurrent downloads if using async download.

    :returns: Tuple where the first list contains names of downloaded datasets and the second list contains
    names of cached datasets
    :rtype: Tuple[List, List]
    """
    project_info = api.project.get_info_by_id(project_id)
    project_meta = ProjectMeta.from_json(api.project.get_meta(project_id))
    if dataset_infos is not None and dataset_ids is not None:
        raise ValueError("dataset_infos and dataset_ids cannot be specified at the same time")
    all_ds_infos = api.dataset.get_list(project_id, recursive=True)
    if dataset_infos is None:
        if dataset_ids is None:
            dataset_infos = all_ds_infos
        else:
            dataset_infos = [ds_info for ds_info in all_ds_infos if ds_info.id in dataset_ids]
    path_to_info = {get_dataset_path(api, dataset_infos, info.id): info for info in dataset_infos}
    to_download, cached = _validate(api, project_info, project_meta, dataset_infos, all_ds_infos)
    if progress_cb is not None:
        cached_items_n = sum(path_to_info[ds_path].items_count for ds_path in cached)
        progress_cb(cached_items_n)
    _download_project_to_cache(
        api=api,
        project_info=project_info,
        dataset_infos=[path_to_info[ds_path] for ds_path in to_download],
        log_progress=log_progress,
        progress_cb=progress_cb,
        semaphore=semaphore,
        **kwargs,
    )
    return to_download, cached


def _get_dataset_parents(api: Api, dataset_infos: List[DatasetInfo], dataset_id):
    dataset_infos_dict = {info.id: info for info in dataset_infos}
    this_dataset_info = dataset_infos_dict.get(dataset_id, None)
    if this_dataset_info is None:
        this_dataset_info = api.dataset.get_info_by_id(dataset_id)
    if this_dataset_info.parent_id is None:
        return []
    parent = _get_dataset_parents(
        api, list(dataset_infos_dict.values()), this_dataset_info.parent_id
    )
    this_parent = dataset_infos_dict.get(this_dataset_info.parent_id, None)
    if this_parent is None:
        this_parent = api.dataset.get_info_by_id(this_dataset_info.parent_id)
    return [*parent, this_parent.name]


def get_dataset_path(api: Api, dataset_infos: List[DatasetInfo], dataset_id: int) -> str:
    parents = _get_dataset_parents(api, dataset_infos, dataset_id)
    dataset_infos_dict = {info.id: info for info in dataset_infos}
    this_dataset_info = dataset_infos_dict.get(dataset_id, None)
    if this_dataset_info is None:
        this_dataset_info = api.dataset.get_info_by_id(dataset_id)
    return Dataset._get_dataset_path(this_dataset_info.name, parents)


def copy_from_cache(
    project_id: int,
    dest_dir: str,
    dataset_names: List[str] = None,
    progress_cb: Callable = None,
    dataset_paths: List[str] = None,
):
    """
    Copy project or dataset from cache to the specified directory.
    If dataset_name is None, the whole project will be copied.

    :param project_id: Project ID, which will be downloaded.
    :type project_id: int
    :param dest_dir: Destination path to local directory.
    :type dest_dir: str
    :param dataset_name: List of dataset paths to copy. If not specified, the whole project will be copied.
    :type dataset_name: str, optional
    :param progress_cb: Function for tracking copying progress. Will be called with number of bytes copied.
    :type progress_cb: tqdm or callable, optional
    :param dataset_paths: List of dataset paths to copy. If not specified, all datasets will be copied.
    :type dataset_paths: list(str), optional

    :returns: None.
    :rtype: NoneType
    """
    if not is_cached(project_id):
        raise RuntimeError(f"Project {project_id} is not cached")
    if dataset_names is not None or dataset_paths is not None:
        if dataset_names is not None:
            dataset_paths = dataset_names
        for dataset_path in dataset_paths:
            if not is_cached(project_id, dataset_path):
                raise RuntimeError(f"Dataset {dataset_path} of project {project_id} is not cached")
    cache_dir = _get_cache_dir(project_id)
    if dataset_paths is None:
        copy_dir_recursively(cache_dir, dest_dir, progress_cb)
    else:
        # copy meta
        copy_file(os.path.join(cache_dir, "meta.json"), os.path.join(dest_dir, "meta.json"))
        # copy datasets
        for dataset_path in dataset_paths:
            copy_dir_recursively(
                os.path.join(cache_dir, dataset_path),
                os.path.join(dest_dir, dataset_path),
                progress_cb,
            )


def download_using_cache(
    api: Api,
    project_id: int,
    dest_dir: str,
    dataset_ids: Optional[List[int]] = None,
    log_progress: bool = True,
    progress_cb: Optional[Union[tqdm, Callable]] = None,
    semaphore: Optional[asyncio.Semaphore] = None,
    **kwargs,
) -> None:
    """
    Download project to the specified directory using cache.
    If dataset_ids is None, all datasets of the project will be downloaded.

    :param api: Supervisely API address and token.
    :type api: :class:`~supervisely.api.api.Api`
    :param project_id: Project ID, which will be downloaded.
    :type project_id: int
    :param dest_dir: Destination path to local directory.
    :type dest_dir: str
    :param dataset_ids: Specified list of Dataset IDs which will be downloaded.
    :type dataset_ids: list(int), optional
    :param log_progress: Show downloading logs in the output.
    :type log_progress: bool
    :param progress_cb: Function for tracking download progress. Will be called with number of items downloaded.
    :type progress_cb: tqdm or callable, optional
    :param semaphore: Semaphore for limiting the number of concurrent downloads if using async download.
    :type semaphore: asyncio.Semaphore, optional

    :returns: None.
    :rtype: NoneType
    """
    downloaded, cached = download_to_cache(
        api,
        project_id,
        dataset_ids=dataset_ids,
        log_progress=log_progress,
        progress_cb=progress_cb,
        semaphore=semaphore,
        **kwargs,
    )
    copy_from_cache(project_id, dest_dir, [*downloaded, *cached])


def read_from_cached_project(
    project_id: int, dataset_name: str, image_names: List[int]
) -> List[Tuple[str, str]]:
    """
    Read images from cached project.

    :param project_id: Project ID.
    :type project_id: int
    :param dataset_name: Name of the dataset.
    :type dataset_name: str
    :param image_ids: List of image IDs.
    :type image_ids: list(int)

    :returns: List of tuples of image path and annotation path.
    :rtype: list(str)
    """
    if not is_cached(project_id, dataset_name):
        raise RuntimeError(f"Dataset {dataset_name} of project {project_id} is not cached")

    dataset = Dataset(_get_cache_dir(project_id, dataset_name), OpenMode.READ)
    paths = []
    for image_name in image_names:
        image_path, ann_path = dataset.get_item_paths(image_name)
        paths.append((image_path, ann_path))
    return paths



from __future__ import annotations

from supervisely.nn.benchmark.object_detection.base_vis_metric import DetectionVisMetric
from supervisely.nn.benchmark.visualization.widgets import ChartWidget, MarkdownWidget


class PerClassAvgPrecision(DetectionVisMetric):
    """Per-class average precision chart for detection reports."""

    MARKDOWN = "per_class_avg_precision"
    CHART = "per_class_avg_precision"

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.clickable = True

    @property
    def md(self) -> MarkdownWidget:
        text = self.vis_texts.markdown_class_ap
        text = text.format(self.vis_texts.definitions.average_precision)
        return MarkdownWidget(self.MARKDOWN, "Average Precision by Class", text)

    @property
    def chart(self) -> ChartWidget:
        chart = ChartWidget(self.CHART, self._get_figure())
        chart.set_click_data(
            self.explore_modal_table.id,
            self.get_click_data(),
            chart_click_extra="'getKey': (payload) => `${payload.points[0].theta}`,",
        )
        return chart

    def _get_figure(self):  # -> go.Figure:
        import plotly.express as px  # pylint: disable=import-error

        # AP per-class
        ap_per_class = self.eval_result.mp.AP_per_class()
        labels = dict(r="Average Precision", theta="Class")
        fig = px.scatter_polar(
            r=ap_per_class,
            theta=self.eval_result.mp.cat_names,
            # title="Per-class Average Precision (AP)",
            labels=labels,
            width=800,
            height=800,
            range_r=[0, 1],
        )
        fig.update_traces(fill="toself")
        fig.update_layout(
            modebar_add=["resetScale"],
            margin=dict(l=80, r=80, t=0, b=0),
        )
        fig.update_traces(
            hovertemplate=labels["theta"]
            + ": %{theta}<br>"
            + labels["r"]
            + ": %{r:.2f}<br>"
            + "<extra></extra>"
        )
        return fig



"""
Tests for FastTable functionality.
Tests also verify that duplicate column names are handled correctly.
"""

import pandas as pd
import pytest

from supervisely.app.widgets import FastTable


class TestFastTableMultiIndex:
    """Test suite for FastTable with duplicate column names (MultiIndex)."""

    def setup_method(self):
        """Setup test data with duplicate column names."""
        # Create data with duplicate column names
        self.duplicate_columns = ["apple", "banana", "apple", "orange"]
        self.test_data = [
            [1, 2, 3, 4],
            [5, 6, 7, 8],
            [9, 10, 11, 12],
        ]
        
        # Create DataFrame with duplicate columns
        self.df = pd.DataFrame(self.test_data, columns=self.duplicate_columns)

    def test_multiindex_creation(self):
        """Test that MultiIndex is created for duplicate column names."""
        table = FastTable(data=self.df)
        
        # Check that internal data has MultiIndex
        assert isinstance(table._source_data.columns, pd.MultiIndex)
        assert table._source_data.columns.names == ["first", "second"]
        
        # Check first level (column names)
        first_level = table._source_data.columns.get_level_values("first").tolist()
        assert first_level == self.duplicate_columns
        
        # Check second level (unique indices)
        second_level = table._source_data.columns.get_level_values("second").tolist()
        assert second_level == [0, 1, 2, 3]

    def test_multiindex_with_list_input(self):
        """Test MultiIndex creation when passing list data with columns parameter."""
        table = FastTable(data=self.test_data, columns=self.duplicate_columns)
        
        assert isinstance(table._source_data.columns, pd.MultiIndex)
        first_level = table._source_data.columns.get_level_values("first").tolist()
        assert first_level == self.duplicate_columns

    def test_output_columns_simple_names(self):
        """Test that output columns are simple strings, not tuples."""
        table = FastTable(data=self.df)
        
        json_data = table.get_json_data()
        columns = json_data["columns"]
        
        # Columns should be simple strings (first level only)
        assert columns == self.duplicate_columns
        
        # Ensure no tuples in columns
        for col in columns:
            assert isinstance(col, str)
            assert not isinstance(col, tuple)

    def test_sorting_with_duplicate_columns(self):
        """Test that sorting works correctly with duplicate column names."""
        table = FastTable(data=self.df, sort_column_idx=0, sort_order="desc")
        
        # Get sorted data
        sorted_df = table._sorted_data
        
        # Check that first column ('apple', 0) is sorted descending
        first_col_values = sorted_df.iloc[:, 0].tolist()
        assert first_col_values == [9, 5, 1]

    def test_sorting_second_duplicate_column(self):
        """Test sorting by the second occurrence of duplicate column name."""
        table = FastTable(data=self.df)
        
        # Sort by column index 2 (second 'apple' column)
        table.sort(column_idx=2, order="asc")
        
        sorted_df = table._sorted_data
        third_col_values = sorted_df.iloc[:, 2].tolist()
        assert third_col_values == [3, 7, 11]

    def test_search_with_duplicate_columns(self):
        """Test that search works across all columns including duplicates."""
        table = FastTable(data=self.df)
        
        # Search for value that exists in multiple places
        searched_data = table._search("7")
        
        # Should find the row containing 7
        assert len(searched_data) == 1
        assert searched_data.iloc[0, 1] == 6  # banana column
        assert searched_data.iloc[0, 2] == 7  # second apple column

    def test_cell_value_access_both_duplicates(self):
        """Test that we can access values from both duplicate columns."""
        table = FastTable(data=self.df)
        
        # Access first 'apple' column (index 0)
        first_apple_col = table._source_data.columns[0]
        first_values = table._source_data[first_apple_col].tolist()
        assert first_values == [1, 5, 9]
        
        # Access second 'apple' column (index 2)
        second_apple_col = table._source_data.columns[2]
        second_values = table._source_data[second_apple_col].tolist()
        assert second_values == [3, 7, 11]
        
        # Verify they are different
        assert first_values != second_values

    def test_update_cell_with_duplicate_columns(self):
        """Test updating cells in duplicate columns."""
        table = FastTable(data=self.df)
        
        # Update first 'apple' column
        table.update_cell_value(row=0, column=0, value=999)
        assert table._source_data.iloc[0, 0] == 999
        
        # Update second 'apple' column
        table.update_cell_value(row=0, column=2, value=888)
        assert table._source_data.iloc[0, 2] == 888
        
        # Verify they are independent
        assert table._source_data.iloc[0, 0] == 999
        assert table._source_data.iloc[0, 2] == 888

    def test_insert_row_with_duplicate_columns(self):
        """Test inserting rows when columns have duplicates."""
        table = FastTable(data=self.df)
        
        new_row = [100, 200, 300, 400]
        table.insert_row(new_row, index=1)
        
        # Verify row was inserted
        assert len(table._source_data) == 4
        assert table._source_data.iloc[1, 0] == 100
        assert table._source_data.iloc[1, 2] == 300  # second apple column

    def test_to_pandas_preserves_column_names(self):
        """Test that to_pandas returns DataFrame with original column names."""
        table = FastTable(data=self.df)
        
        exported_df = table.to_pandas()
        
        # Should have original duplicate column names
        assert exported_df.columns.tolist() == self.duplicate_columns
        
        # Data should be preserved
        assert exported_df.iloc[0].tolist() == self.test_data[0]

    def test_to_json_simple_column_names(self):
        """Test that to_json exports simple column names."""
        table = FastTable(data=self.df)
        
        json_export = table.to_json()
        
        # Columns should be simple strings
        assert json_export["columns"] == self.duplicate_columns
        
        # Data should be preserved
        assert json_export["data"][0] == self.test_data[0]

    def test_select_row_by_value_duplicate_columns(self):
        """Test selecting row by value when column names are duplicated."""
        table = FastTable(data=self.df, is_selectable=True)
        
        # Select by first 'apple' column - should work with column name
        # Note: When there are duplicates, this will match the first occurrence
        table.select_row_by_value("apple", 5)
        
        selected = table.get_selected_row()
        assert selected.row_index == 1
        assert selected.row == [5, 6, 7, 8]

    def test_multiindex_with_numeric_data(self):
        """Test MultiIndex with numeric sorting."""
        # Create data where numeric sorting matters
        numeric_data = [
            ["3", "a", "29", "x"],
            ["29", "b", "3", "y"],
            ["100", "c", "10", "z"],
        ]
        
        table = FastTable(
            data=numeric_data, 
            columns=["value", "letter", "value", "char"],
            sort_column_idx=0,
            sort_order="asc"
        )
        
        # Should sort numerically: 3, 29, 100
        sorted_df = table._sorted_data
        first_col_values = sorted_df.iloc[:, 0].tolist()
        assert first_col_values == ["3", "29", "100"]

    def test_multiindex_filter_and_search(self):
        """Test that filter and search work together with MultiIndex."""
        table = FastTable(data=self.df, page_size=2)
        
        # Search should work
        table.search("1")
        
        # Should find rows with '1' in any column
        assert table._rows_total >= 1

    def test_no_multiindex_for_unique_columns(self):
        """Test that MultiIndex is NOT created when all column names are unique."""
        unique_columns = ["apple", "banana", "orange", "grape"]
        df_unique = pd.DataFrame(self.test_data, columns=unique_columns)
        
        table = FastTable(data=df_unique)
        
        # Should have MultiIndex (always created now for consistency)
        # But it should handle unique names correctly
        first_level = table._source_data.columns.get_level_values("first").tolist()
        assert first_level == unique_columns

    def test_multiindex_with_none_and_nan(self):
        """Test MultiIndex with None and NaN values in duplicate columns."""
        data_with_none = [
            [1, None, 3, 4],
            [5, 6, None, 8],
            [None, 10, 11, None],
        ]
        
        table = FastTable(data=data_with_none, columns=self.duplicate_columns)
        
        # Should convert None to ""
        json_data = table.get_json_data()
        
        # Check that None was converted to empty string
        assert json_data["data"][0]["items"][1] == ""
        assert json_data["data"][1]["items"][2] == ""

    def test_multiindex_pagination(self):
        """Test pagination works correctly with MultiIndex."""
        # Create larger dataset
        large_data = [[i, i+1, i+2, i+3] for i in range(25)]
        
        table = FastTable(
            data=large_data, 
            columns=self.duplicate_columns,
            page_size=10
        )
        
        # Check first page
        assert len(table._parsed_active_data["data"]) == 10
        
        # Check total rows
        assert table._rows_total == 25

    def test_multiindex_column_access_by_tuple(self):
        """Test that internal DataFrame allows tuple column access."""
        table = FastTable(data=self.df)
        
        # Access by tuple should work
        first_apple = table._source_data[("apple", 0)]
        second_apple = table._source_data[("apple", 2)]
        
        assert first_apple.tolist() == [1, 5, 9]
        assert second_apple.tolist() == [3, 7, 11]

    def test_multiindex_prevents_column_collision_error(self):
        """Test that MultiIndex prevents errors from duplicate column names."""
        # Without MultiIndex, accessing duplicate columns would cause issues
        # With MultiIndex, each column is uniquely accessible
        
        table = FastTable(data=self.df)
        
        # This should work without errors
        all_columns = []
        for i in range(len(table._source_data.columns)):
            col = table._source_data.columns[i]
            values = table._source_data[col].tolist()
            all_columns.append(values)
        
        # Should have 4 columns
        assert len(all_columns) == 4
        
        # Two 'apple' columns should have different values
        assert all_columns[0] != all_columns[2]


class TestFastTableDataFormats:
    """Test suite for different data input formats."""

    def test_empty_table_with_none(self):
        """Test creating empty table with None data."""
        table = FastTable(data=None, columns=["col1", "col2", "col3"])
        
        assert len(table._source_data) == 0
        assert table._rows_total == 0
        assert table._columns_first_idx == ["col1", "col2", "col3"]

    def test_empty_table_without_data(self):
        """Test creating empty table without data parameter."""
        table = FastTable(columns=["a", "b", "c"])
        
        assert len(table._source_data) == 0
        assert table._rows_total == 0

    def test_empty_dataframe(self):
        """Test creating table with empty DataFrame."""
        df = pd.DataFrame(columns=["col1", "col2", "col3"])
        table = FastTable(data=df)
        
        assert len(table._source_data) == 0
        assert table._rows_total == 0
        assert table._columns_first_idx == ["col1", "col2", "col3"]

    def test_list_format_basic(self):
        """Test creating table with basic list format."""
        data = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        columns = ["a", "b", "c"]
        
        table = FastTable(data=data, columns=columns)
        
        assert table._rows_total == 3
        assert table._columns_first_idx == columns
        
        # Check data is preserved
        exported = table.to_pandas()
        assert exported.iloc[0].tolist() == [1, 2, 3]
        assert exported.iloc[2].tolist() == [7, 8, 9]

    def test_list_format_single_row(self):
        """Test creating table with single row in list format."""
        data = [[100, 200, 300]]
        columns = ["x", "y", "z"]
        
        table = FastTable(data=data, columns=columns)
        
        assert table._rows_total == 1
        assert table.to_pandas().iloc[0].tolist() == [100, 200, 300]

    def test_list_format_empty_list(self):
        """Test creating table with empty list."""
        data = []
        columns = ["col1", "col2"]
        
        table = FastTable(data=data, columns=columns)
        
        assert table._rows_total == 0
        assert len(table._source_data) == 0

    def test_list_format_with_none_values(self):
        """Test list data with None values gets converted to empty strings."""
        data = [[1, None, 3], [None, 5, None], [7, 8, 9]]
        columns = ["a", "b", "c"]
        
        table = FastTable(data=data, columns=columns)
        json_data = table.get_json_data()
        
        # None should be converted to ""
        assert json_data["data"][0]["items"][1] == ""
        assert json_data["data"][1]["items"][0] == ""
        assert json_data["data"][1]["items"][2] == ""

    def test_dataframe_with_range_index(self):
        """Test DataFrame with default numeric column names."""
        data = [[1, 2, 3], [4, 5, 6]]
        df = pd.DataFrame(data)  # No column names specified
        
        table = FastTable(data=df)
        
        assert table._rows_total == 2
        # Should have no MultiIndex for RangeIndex
        assert table._columns_first_idx is None or isinstance(table._source_data.columns, pd.RangeIndex)

    def test_read_pandas_replaces_data(self):
        """Test that read_pandas method replaces table data."""
        initial_data = [[1, 2], [3, 4]]
        table = FastTable(data=initial_data, columns=["a", "b"])
        
        assert table._rows_total == 2
        
        # Replace with new data - same columns!
        new_df = pd.DataFrame([[99, 88], [77, 66]], columns=["a", "b"])
        table.read_pandas(new_df)
        
        assert table._rows_total == 2
        # read_pandas keeps original column names
        assert table._columns_first_idx == ["a", "b"]
        exported = table.to_pandas()
        assert exported.iloc[0].tolist() == [99, 88]

    def test_read_json_replaces_data(self):
        """Test that read_json method replaces table data."""
        initial_data = [[1, 2], [3, 4]]
        table = FastTable(data=initial_data, columns=["a", "b"])
        
        # Replace with new data via JSON (must include options for read_json)
        new_data = {
            "data": [[100, 200], [300, 400], [500, 600]],
            "columns": ["col1", "col2"],
            "options": {}  # read_json requires options field
        }
        table.read_json(new_data)
        
        assert table._rows_total == 3
        json_export = table.to_json()
        assert json_export["data"][0] == [100, 200]
        assert json_export["data"][2] == [500, 600]

    def test_clear_method_empties_table(self):
        """Test that clear method removes all data."""
        data = [[1, 2, 3], [4, 5, 6]]
        table = FastTable(data=data, columns=["a", "b", "c"])
        
        assert table._rows_total == 2
        
        table.clear()
        
        assert table._rows_total == 0
        assert len(table._source_data) == 0

    def test_add_rows_method(self):
        """Test adding multiple rows at once."""
        table = FastTable(data=[[1, 2]], columns=["a", "b"])
        
        assert table._rows_total == 1
        
        table.add_rows([[3, 4], [5, 6], [7, 8]])
        
        assert table._rows_total == 4
        exported = table.to_pandas()
        assert exported.iloc[3].tolist() == [7, 8]

    def test_pop_row_removes_and_returns(self):
        """Test pop_row method removes and returns a row."""
        data = [[1, 2], [3, 4], [5, 6]]
        table = FastTable(data=data, columns=["a", "b"])
        
        assert table._rows_total == 3
        
        # Pop last row
        popped = table.pop_row(-1)
        
        assert table._rows_total == 2
        assert popped.tolist() == [5, 6]

    def test_mixed_data_types_in_columns(self):
        """Test table with mixed data types (strings and numbers)."""
        data = [
            ["text1", 100, 1.5],
            ["text2", 200, 2.5],
            ["text3", 300, 3.5],
        ]
        columns = ["str_col", "int_col", "float_col"]
        
        table = FastTable(data=data, columns=columns)
        
        assert table._rows_total == 3
        exported = table.to_pandas()
        assert exported.iloc[1, 0] == "text2"
        assert exported.iloc[1, 1] == 200
        assert exported.iloc[1, 2] == 2.5


class TestFastTableEdgeCases:
    """Test suite for edge cases and special scenarios."""

    def test_very_long_table(self):
        """Test table with many rows (stress test)."""
        data = [[i, i*2, i*3] for i in range(1000)]
        columns = ["x", "y", "z"]

        table = FastTable(data=data, columns=columns, page_size=50)

        assert table._rows_total == 1000
        assert len(table._parsed_active_data["data"]) == 50  # First page

    def test_wide_table(self):
        """Test table with many columns."""
        num_cols = 50
        columns = [f"col{i}" for i in range(num_cols)]
        data = [[i for i in range(num_cols)]]

        table = FastTable(data=data, columns=columns)

        assert table._rows_total == 1
        assert len(table._columns_first_idx) == num_cols

    def test_special_characters_in_column_names(self):
        """Test column names with special characters."""
        columns = ["col_1", "col-2", "col.3", "col@4", "col 5"]
        data = [[1, 2, 3, 4, 5]]

        table = FastTable(data=data, columns=columns)

        assert table._columns_first_idx == columns
        assert table._rows_total == 1

    def test_unicode_in_data(self):
        """Test table with unicode characters."""
        data = [
            ["こんにちは", "你好", "مرحبا", "Hello"],
            ["さようなら", "再见", "مع السلامة", "Goodbye 👋"],
        ]
        columns = ["Japanese", "Chinese", "Arabic", "English"]

        table = FastTable(data=data, columns=columns)

        assert table._rows_total == 2
        exported = table.to_pandas()
        assert exported.iloc[0, 0] == "こんにちは"
        assert exported.iloc[1, 3] == "Goodbye 👋"

    def test_duplicate_column_with_select_rows_by_value(self):
        """Test select_rows_by_value with duplicate columns."""
        data = [[1, 2, 3], [5, 6, 7], [5, 8, 9]]
        columns = ["apple", "banana", "apple"]

        table = FastTable(data=data, columns=columns, is_selectable=True)

        # Select rows where first 'apple' column has value 5
        table.select_rows_by_value("apple", [5])

        selected = table.get_selected_rows()
        assert len(selected) == 2
        assert selected[0].row_index == 1
        assert selected[1].row_index == 2

    def test_three_duplicate_columns(self):
        """Test table with three columns having the same name."""
        data = [[1, 2, 3, 4], [5, 6, 7, 8]]
        columns = ["value", "value", "value", "other"]

        table = FastTable(data=data, columns=columns)

        # Should have MultiIndex
        assert isinstance(table._source_data.columns, pd.MultiIndex)

        # All three 'value' columns should be accessible
        first_value = table._source_data[("value", 0)].tolist()
        second_value = table._source_data[("value", 1)].tolist()
        third_value = table._source_data[("value", 2)].tolist()

        assert first_value == [1, 5]
        assert second_value == [2, 6]
        assert third_value == [3, 7]


class TestFastTableNoneHandling:
    """Test suite for proper None/NaN handling in FastTable."""

    def setup_method(self):
        """Setup test data with None values."""
        self.data_with_none = [
            [1, None, 3, "text"],
            [None, 6, None, "value"],
            [9, 10, 11, None],
        ]
        self.columns = ["col1", "col2", "col3", "col4"]

    def test_none_values_preserved_in_source_data(self):
        """Test that None values are preserved in _source_data (as NaN in pandas)."""
        table = FastTable(data=self.data_with_none, columns=self.columns)

        # None is converted to NaN by pandas, which should be preserved in source data
        assert pd.isna(table._source_data.iloc[0, 1])
        assert pd.isna(table._source_data.iloc[1, 0])
        assert pd.isna(table._source_data.iloc[1, 2])
        assert pd.isna(table._source_data.iloc[2, 3])

    def test_none_values_converted_to_empty_string_in_json(self):
        """Test that None values are converted to empty strings in JSON output."""
        table = FastTable(data=self.data_with_none, columns=self.columns)

        json_data = table.get_json_data()

        # None should be converted to "" for display
        assert json_data["data"][0]["items"][1] == ""
        assert json_data["data"][1]["items"][0] == ""
        assert json_data["data"][1]["items"][2] == ""
        assert json_data["data"][2]["items"][3] == ""

        # Non-None values should be preserved
        assert json_data["data"][0]["items"][0] == 1
        assert json_data["data"][0]["items"][3] == "text"

    def test_to_pandas_preserves_none_values(self):
        """Test that to_pandas() returns DataFrame with NaN values (pandas behavior)."""
        table = FastTable(data=self.data_with_none, columns=self.columns)

        exported_df = table.to_pandas()

        # pandas converts None to NaN, which should be preserved
        assert pd.isna(exported_df.iloc[0, 1])
        assert pd.isna(exported_df.iloc[1, 0])
        assert pd.isna(exported_df.iloc[1, 2])
        assert pd.isna(exported_df.iloc[2, 3])

    def test_get_clicked_row_returns_none_values(self):
        """Test that get_clicked_row returns NaN values from source (not empty strings)."""
        table = FastTable(data=self.data_with_none, columns=self.columns)

        # Directly get data from source for row 1
        row_data = table._source_data.loc[1].values.tolist()

        # Row should have NaN values (pandas converts None to NaN)
        assert pd.isna(row_data[0])  # col1
        assert pd.isna(row_data[2])  # col3
        assert row_data[1] == 6  # col2
        assert row_data[3] == "value"  # col4

    def test_get_clicked_cell_returns_none_value(self):
        """Test that get_clicked_cell returns NaN value from source."""
        import json

        from supervisely.app.content import StateJson

        table = FastTable(data=self.data_with_none, columns=self.columns)

        # Manually set clicked cell state to simulate cell click
        StateJson()[table.widget_id]["selectedCell"] = {
            "idx": 0,
            "column": 1,
            "row": [1, None, 3, "text"],  # This gets ignored, we read from source
        }

        clicked_cell = table.get_clicked_cell()

        # Should return NaN from source data (pandas converts None to NaN)
        assert pd.isna(clicked_cell.column_value)
        assert clicked_cell.column_name == "col2"
        assert clicked_cell.row_index == 0

    def test_select_row_by_value_with_none(self):
        """Test working with NaN values in source data."""
        import numpy as np

        table = FastTable(data=self.data_with_none, columns=self.columns, is_selectable=True)

        # pandas converts None to NaN
        # Verify that NaN values exist in source
        col_idx = table._columns_first_idx.index("col2")
        col_tuple = table._source_data.columns[col_idx]
        nan_rows = table._source_data[table._source_data[col_tuple].isna()]

        assert len(nan_rows) > 0, "Should have NaN values in col2"
        assert pd.isna(table._source_data.iloc[0, 1]), "First row col2 should be NaN"

    def test_select_rows_by_value_with_none(self):
        """Test searching for NaN values using pandas isna()."""
        table = FastTable(data=self.data_with_none, columns=self.columns, is_selectable=True)

        # Find rows where col3 is NaN using pandas
        col_idx = table._columns_first_idx.index("col3")
        col_tuple = table._source_data.columns[col_idx]
        nan_rows = table._source_data[table._source_data[col_tuple].isna()]

        # Select those rows manually
        indices = nan_rows.index.tolist()
        table.select_rows(indices)

        selected = table.get_selected_rows()
        assert len(selected) == 1
        assert selected[0].row_index == 1
        assert pd.isna(selected[0].row[2])

    def test_pop_row_returns_none_values(self):
        """Test that pop_row returns row with NaN values."""
        table = FastTable(data=self.data_with_none, columns=self.columns)

        popped = table.pop_row(0)

        # pandas converts None to NaN
        assert pd.isna(popped[1])
        assert popped[0] == 1
        assert popped[2] == 3

    def test_update_cell_with_none(self):
        """Test updating a cell to None value (becomes NaN in pandas)."""
        table = FastTable(data=[[1, 2, 3]], columns=["a", "b", "c"])

        # Update cell to None
        table.update_cell_value(row=0, column=1, value=None)

        # Source data should have NaN (pandas converts None to NaN)
        assert pd.isna(table._source_data.iloc[0, 1])

        # JSON output should have ""
        json_data = table.get_json_data()
        assert json_data["data"][0]["items"][1] == ""

        # to_pandas should have NaN
        exported = table.to_pandas()
        assert pd.isna(exported.iloc[0, 1])

    def test_insert_row_with_none_values(self):
        """Test inserting a row containing None values."""
        table = FastTable(data=[[1, 2, 3]], columns=["a", "b", "c"])

        new_row = [None, 5, None]
        table.insert_row(new_row, index=0)

        # Source data should preserve None
        assert table._source_data.iloc[0, 0] is None
        assert table._source_data.iloc[0, 2] is None
        assert table._source_data.iloc[0, 1] == 5

    def test_add_rows_with_none_values(self):
        """Test adding multiple rows with None values (become NaN in pandas)."""
        table = FastTable(data=[[1, 2]], columns=["a", "b"])

        table.add_rows([[None, 4], [5, None], [None, None]])

        # Verify None values are preserved as NaN
        assert pd.isna(table._source_data.iloc[1, 0])
        assert pd.isna(table._source_data.iloc[2, 1])
        assert pd.isna(table._source_data.iloc[3, 0])
        assert pd.isna(table._source_data.iloc[3, 1])

    def test_search_with_none_values(self):
        """Test that search works correctly with NaN values."""
        table = FastTable(data=self.data_with_none, columns=self.columns)

        # Search for "6" should find row with NaN values
        searched = table._search("6")

        assert len(searched) == 1
        assert pd.isna(searched.iloc[0, 0])  # NaN is preserved
        assert searched.iloc[0, 1] == 6

    def test_sort_with_none_values(self):
        """Test that sorting works with None values."""
        data = [[3, None], [1, "b"], [None, "a"], [2, "c"]]
        table = FastTable(data=data, columns=["num", "letter"], sort_column_idx=0, sort_order="asc")

        sorted_df = table._sorted_data

        # None values should be handled during sort
        assert sorted_df is not None
        assert len(sorted_df) == 4

    def test_filter_with_none_values(self):
        """Test custom filter function with None values."""
        table = FastTable(data=self.data_with_none, columns=self.columns)

        def filter_non_none_in_col2(data, value):
            # Filter rows where col2 is not None
            col_tuple = data.columns[1]  # col2
            return data[data[col_tuple].notna()]

        table.set_filter(filter_non_none_in_col2)
        table.filter(True)

        # Should only have rows where col2 is not None
        assert table._rows_total == 2  # rows 1 and 2 (indices 1 and 2)

    def test_dataframe_with_nan_values(self):
        """Test that pandas NaN values are also handled correctly."""
        import numpy as np

        df = pd.DataFrame({"a": [1, np.nan, 3], "b": [np.nan, 2, 3], "c": [1, 2, np.nan]})

        table = FastTable(data=df)

        # NaN should be converted to "" in JSON
        json_data = table.get_json_data()
        assert json_data["data"][0]["items"][1] == ""
        assert json_data["data"][1]["items"][0] == ""
        assert json_data["data"][2]["items"][2] == ""

        # But preserved in to_pandas
        exported = table.to_pandas()
        assert pd.isna(exported.iloc[0, 1])
        assert pd.isna(exported.iloc[1, 0])
        assert pd.isna(exported.iloc[2, 2])

    def test_to_json_converts_none_to_empty_string(self):
        """Test that to_json exports None as empty strings."""
        table = FastTable(data=self.data_with_none, columns=self.columns)

        json_export = table.to_json()

        # None should be "" in JSON export
        assert json_export["data"][0][1] == ""
        assert json_export["data"][1][0] == ""
        assert json_export["data"][1][2] == ""

    def test_active_page_with_none_preserves_values(self):
        """Test that to_pandas(active_page=True) preserves NaN values."""
        table = FastTable(data=self.data_with_none, columns=self.columns, page_size=2)

        # Get active page
        exported = table.to_pandas(active_page=True)

        # Should preserve NaN values from active page
        assert pd.isna(exported.iloc[0, 1])
        assert pd.isna(exported.iloc[1, 0])

    def test_multiindex_with_none_values(self):
        """Test None handling with duplicate column names (MultiIndex)."""
        data = [[1, None, 3], [None, 5, None]]
        columns = ["value", "value", "other"]

        table = FastTable(data=data, columns=columns)

        # Verify MultiIndex is created
        assert isinstance(table._source_data.columns, pd.MultiIndex)

        # None is converted to NaN by pandas and should be preserved in source
        assert pd.isna(table._source_data.iloc[0, 1])
        assert pd.isna(table._source_data.iloc[1, 0])
        assert pd.isna(table._source_data.iloc[1, 2])

        # to_pandas should preserve NaN
        exported = table.to_pandas()
        assert pd.isna(exported.iloc[0, 1])
        assert pd.isna(exported.iloc[1, 0])


class TestFastTableSorting:
    """Comprehensive test suite for FastTable sorting functionality."""

    def test_sort_numeric_ascending(self):
        """Test sorting numeric column in ascending order."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 0].tolist() == [1, 2, 5, 8]
        # Verify data is synchronized
        assert table._rows_total == 4
        assert len(table._parsed_active_data["data"]) == 4

    def test_sort_numeric_descending(self):
        """Test sorting numeric column in descending order."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="desc")

        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 0].tolist() == [8, 5, 2, 1]

    def test_sort_string_ascending(self):
        """Test sorting string column in ascending order."""
        data = [["zebra", 1], ["apple", 2], ["mango", 3], ["banana", 4]]
        table = FastTable(data=data, columns=["fruit", "count"])

        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 0].tolist() == ["apple", "banana", "mango", "zebra"]

    def test_sort_string_descending(self):
        """Test sorting string column in descending order."""
        data = [["zebra", 1], ["apple", 2], ["mango", 3], ["banana", 4]]
        table = FastTable(data=data, columns=["fruit", "count"])

        table.sort(column_idx=0, order="desc")

        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 0].tolist() == ["zebra", "mango", "banana", "apple"]

    def test_sort_numeric_strings_as_numbers(self):
        """Test that numeric strings are sorted numerically, not alphabetically."""
        data = [["100", "a"], ["3", "b"], ["29", "c"], ["5", "d"]]
        table = FastTable(data=data, columns=["num_str", "letter"])

        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        # Should sort as numbers: 3, 5, 29, 100 (not "100", "29", "3", "5")
        assert sorted_df.iloc[:, 0].tolist() == ["3", "5", "29", "100"]

    def test_sort_mixed_numeric_strings_descending(self):
        """Test numeric strings sorted descending."""
        data = [["100", "a"], ["3", "b"], ["29", "c"], ["5", "d"]]
        table = FastTable(data=data, columns=["num_str", "letter"])

        table.sort(column_idx=0, order="desc")

        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 0].tolist() == ["100", "29", "5", "3"]

    def test_sort_with_none_values_ascending(self):
        """Test that None values are placed at the end when sorting ascending."""
        data = [[3, "c"], [None, "none1"], [1, "a"], [None, "none2"], [2, "b"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        # None values should be at the end
        values = sorted_df.iloc[:, 0].tolist()
        assert values[:3] == [1, 2, 3]
        assert pd.isna(values[3])
        assert pd.isna(values[4])

    def test_sort_with_none_values_descending(self):
        """Test that None values are placed at the end when sorting descending."""
        data = [[3, "c"], [None, "none1"], [1, "a"], [None, "none2"], [2, "b"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="desc")

        sorted_df = table._sorted_data
        values = sorted_df.iloc[:, 0].tolist()
        assert values[:3] == [3, 2, 1]
        assert pd.isna(values[3])
        assert pd.isna(values[4])

    def test_sort_with_nan_values(self):
        """Test sorting with NaN values from pandas."""
        import numpy as np

        df = pd.DataFrame({"num": [3, np.nan, 1, np.nan, 2], "letter": ["c", "n1", "a", "n2", "b"]})
        table = FastTable(data=df)

        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        values = sorted_df.iloc[:, 0].tolist()
        assert values[:3] == [1, 2, 3]
        assert pd.isna(values[3])
        assert pd.isna(values[4])

    def test_sort_initialization_with_params(self):
        """Test that table is sorted correctly when sort params are passed to constructor."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"], sort_column_idx=0, sort_order="asc")

        # Should be sorted on initialization
        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 0].tolist() == [1, 2, 5, 8]

    def test_sort_none_column_idx(self):
        """Test that passing None parameters without reset preserves current sorting."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        # First sort
        table.sort(column_idx=0, order="asc")
        assert table._sorted_data.iloc[:, 0].tolist() == [1, 2, 5, 8]

        # Call with None params (without reset) - should preserve current sorting
        table.sort(column_idx=None, order=None)

        # Should keep current sorting
        assert table._sorted_data.iloc[:, 0].tolist() == [1, 2, 5, 8]
        assert table._sort_column_idx == 0
        assert table._sort_order == "asc"

    def test_sort_none_order(self):
        """Test that passing only order=None keeps current column."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        # Don't sort on initialization
        table = FastTable(data=data, columns=["num", "letter"])

        # Sort the table first
        table.sort(column_idx=0, order="asc")
        assert table._sorted_data.iloc[:, 0].tolist() == [1, 2, 5, 8]
        assert table._sort_column_idx == 0
        assert table._sort_order == "asc"

        # Change only order, keep column
        table.sort(order="desc")
        assert table._sorted_data.iloc[:, 0].tolist() == [8, 5, 2, 1]
        assert table._sort_column_idx == 0  # Column preserved
        assert table._sort_order == "desc"

    def test_sort_none_column_keeps_order(self):
        """Test that passing only column_idx=None keeps current order."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        # Sort by column 0 ascending
        table.sort(column_idx=0, order="asc")
        assert table._sort_column_idx == 0
        assert table._sort_order == "asc"

        # Change to column 1, keep order
        table.sort(column_idx=1)
        assert table._sorted_data.iloc[:, 1].tolist() == ["a", "b", "e", "h"]
        assert table._sort_column_idx == 1
        assert table._sort_order == "asc"  # Order preserved

    def test_sort_reset_clears_sorting(self):
        """Test that reset=True clears sorting completely."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        # Sort the table first
        table.sort(column_idx=0, order="asc")
        assert table._sorted_data.iloc[:, 0].tolist() == [1, 2, 5, 8]
        assert table._sort_column_idx == 0
        assert table._sort_order == "asc"

        # Clear sorting with reset=True
        table.sort(reset=True)

        # Should return to original order
        assert table._sorted_data.iloc[:, 0].tolist() == [5, 2, 8, 1]
        assert table._sort_column_idx is None
        assert table._sort_order is None

    def test_sort_no_params_preserves_state(self):
        """Test that sort() without parameters preserves current sorting state."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        # Sort the table first
        table.sort(column_idx=0, order="asc")
        assert table._sort_column_idx == 0
        assert table._sort_order == "asc"

        # Call sort without params - should preserve state
        table.sort()
        assert table._sort_column_idx == 0
        assert table._sort_order == "asc"

    def test_sort_invalid_column_idx(self):
        """Test that invalid column index raises an error."""
        data = [[5, "e"], [2, "b"]]
        table = FastTable(data=data, columns=["num", "letter"])

        with pytest.raises(IndexError) as exc_info:
            table.sort(column_idx=10, order="asc")

        assert "column idx = 10 is not possible" in str(exc_info.value)

    def test_sort_negative_column_idx_validation(self):
        """Test that negative column index is validated and set to None."""
        data = [[5, "e"], [2, "b"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=-1, order="asc")

        # Should be set to None and no sorting applied
        assert table._sort_column_idx is None
        assert table._sorted_data.iloc[:, 0].tolist() == [5, 2]

    def test_sort_invalid_order_validation(self):
        """Test that invalid order value is validated and set to None."""
        data = [[5, "e"], [2, "b"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="invalid")

        # Should be set to None and no sorting applied
        assert table._sort_order is None
        assert table._sorted_data.iloc[:, 0].tolist() == [5, 2]

    def test_sort_with_pagination(self):
        """Test that sorting works correctly with pagination."""
        data = [[i, chr(ord("a") + i)] for i in range(10, 0, -1)]  # 10, 9, ..., 1
        table = FastTable(data=data, columns=["num", "letter"], page_size=3)

        table.sort(column_idx=0, order="asc")

        # First page should have 1, 2, 3
        parsed_data = table._parsed_active_data["data"]
        assert len(parsed_data) == 3
        assert parsed_data[0]["items"][0] == 1
        assert parsed_data[1]["items"][0] == 2
        assert parsed_data[2]["items"][0] == 3

    def test_sort_with_search(self):
        """Test that sorting works correctly after search filtering."""
        data = [[5, "apple"], [2, "banana"], [8, "apricot"], [1, "avocado"], [3, "berry"]]
        table = FastTable(data=data, columns=["num", "fruit"])

        # Search for fruits starting with 'a'
        table.search("a")

        # Sort the filtered results
        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        # Should have only filtered items, sorted: 1, 5, 8
        assert len(sorted_df) == 4  # apple, apricot, avocado, banana
        assert sorted_df.iloc[0, 0] == 1  # avocado

    def test_sort_with_filter(self):
        """Test that sorting works correctly after custom filtering."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"], [6, "f"]]
        table = FastTable(data=data, columns=["num", "letter"])

        # Set custom filter to keep only values > 3
        def filter_gt_3(df, value):
            if value is None:
                return df
            col_tuple = df.columns[0]
            return df[df[col_tuple] > value]

        table.set_filter(filter_gt_3)
        table.filter(3)

        # Sort filtered results
        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        # Should have only 5, 6, 8 sorted
        assert sorted_df.iloc[:, 0].tolist() == [5, 6, 8]

    def test_sort_preserves_row_indices(self):
        """Test that sorting preserves original row indices in parsed data."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        # Check that idx values correspond to original positions
        parsed_data = table._parsed_active_data["data"]
        assert parsed_data[0]["idx"] == 3  # Row with value 1 was at index 3
        assert parsed_data[1]["idx"] == 1  # Row with value 2 was at index 1
        assert parsed_data[2]["idx"] == 0  # Row with value 5 was at index 0
        assert parsed_data[3]["idx"] == 2  # Row with value 8 was at index 2

    def test_sort_updates_state_json(self):
        """Test that sort method updates StateJson correctly."""
        data = [[5, "e"], [2, "b"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=1, order="desc")

        # Verify StateJson is updated
        from supervisely.app.content import StateJson

        assert StateJson()[table.widget_id]["sort"]["column"] == 1
        assert StateJson()[table.widget_id]["sort"]["order"] == "desc"

    def test_sort_updates_data_json(self):
        """Test that sort method updates DataJson with sorted data."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        # Verify DataJson contains sorted data
        from supervisely.app import DataJson

        data_items = DataJson()[table.widget_id]["data"]
        assert data_items[0]["items"][0] == 1
        assert data_items[1]["items"][0] == 2
        assert data_items[2]["items"][0] == 5
        assert data_items[3]["items"][0] == 8

    def test_sort_second_column(self):
        """Test sorting by second column."""
        data = [[1, "z"], [2, "a"], [3, "m"], [4, "b"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=1, order="asc")

        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 1].tolist() == ["a", "b", "m", "z"]

    def test_sort_switch_column(self):
        """Test switching sort column."""
        data = [[5, "e"], [2, "b"], [8, "a"], [1, "z"]]
        table = FastTable(data=data, columns=["num", "letter"])

        # First sort by column 0
        table.sort(column_idx=0, order="asc")
        assert table._sorted_data.iloc[:, 0].tolist() == [1, 2, 5, 8]

        # Then sort by column 1
        table.sort(column_idx=1, order="asc")
        assert table._sorted_data.iloc[:, 1].tolist() == ["a", "b", "e", "z"]

    def test_sort_switch_order(self):
        """Test switching sort order on same column."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        # Sort ascending
        table.sort(column_idx=0, order="asc")
        assert table._sorted_data.iloc[:, 0].tolist() == [1, 2, 5, 8]

        # Sort descending
        table.sort(column_idx=0, order="desc")
        assert table._sorted_data.iloc[:, 0].tolist() == [8, 5, 2, 1]

    def test_sort_empty_table(self):
        """Test sorting an empty table doesn't cause errors."""
        table = FastTable(data=None, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        assert len(table._sorted_data) == 0
        assert table._rows_total == 0

    def test_sort_single_row(self):
        """Test sorting a table with single row."""
        data = [[5, "e"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        assert table._sorted_data.iloc[0, 0] == 5
        assert len(table._sorted_data) == 1

    def test_sort_duplicate_values(self):
        """Test sorting with duplicate values maintains stability."""
        data = [[5, "a"], [5, "b"], [5, "c"], [1, "d"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        # First value should be 1
        assert sorted_df.iloc[0, 0] == 1
        # Rest should be 5's (order may vary)
        assert sorted_df.iloc[1, 0] == 5
        assert sorted_df.iloc[2, 0] == 5
        assert sorted_df.iloc[3, 0] == 5

    def test_sort_mixed_types_column(self):
        """Test sorting column with mixed numeric strings and pure strings."""
        data = [["10", "a"], ["abc", "b"], ["2", "c"], ["xyz", "d"]]
        table = FastTable(data=data, columns=["mixed", "letter"])

        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        # Numeric strings should be sorted first numerically, then non-numeric alphabetically
        values = sorted_df.iloc[:, 0].tolist()
        # Should have: 2, 10, abc, xyz
        assert values == ["2", "10", "abc", "xyz"]

    def test_custom_sort_function(self):
        """Test using a custom sort function."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        # Custom sort: reverse alphabetical on second column
        def custom_sort(data, column_idx, order):
            col = data.columns[column_idx]
            ascending = order == "asc"
            # Reverse the ascending flag for custom behavior
            return data.sort_values(by=col, ascending=not ascending)

        table.set_sort(custom_sort)
        table.sort(column_idx=1, order="asc")

        # With reversed logic, "asc" should give descending
        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 1].tolist() == ["h", "e", "b", "a"]

    def test_sort_after_insert_row(self):
        """Test that sorting works correctly after inserting a row."""
        data = [[5, "e"], [2, "b"], [8, "h"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        # Insert new row
        table.insert_row([1, "a"], index=0)

        # Re-sort
        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        assert sorted_df.iloc[:, 0].tolist() == [1, 2, 5, 8]

    def test_sort_after_pop_row(self):
        """Test that sorting works correctly after removing a row."""
        data = [[5, "e"], [2, "b"], [8, "h"], [1, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        # Remove a row
        table.pop_row(0)

        # Re-sort
        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        # Should have 3 rows now
        assert len(sorted_df) == 3

    def test_sort_with_all_none_column(self):
        """Test sorting a column that contains only None values."""
        data = [[None, "a"], [None, "b"], [None, "c"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        # Should not crash, all values are None
        sorted_df = table._sorted_data
        assert len(sorted_df) == 3
        assert all(pd.isna(sorted_df.iloc[:, 0]))

    def test_sort_preserves_data_types(self):
        """Test that sorting doesn't change data types."""
        data = [[5.5, "e"], [2.1, "b"], [8.9, "h"], [1.0, "a"]]
        table = FastTable(data=data, columns=["num", "letter"])

        table.sort(column_idx=0, order="asc")

        sorted_df = table._sorted_data
        # Verify floats are preserved
        assert sorted_df.iloc[0, 0] == 1.0
        assert isinstance(sorted_df.iloc[0, 0], float)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



from typing import Optional

from supervisely.app.widgets import Widget


class Empty(Widget):
    """Placeholder for empty space in layout."""

    def __init__(self, widget_id: Optional[str] = None, style: Optional[str] = ""):
        """
        :param widget_id: Widget identifier.
        :type widget_id: Optional[str]
        :param style: CSS styles.
        :type style: Optional[str]

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import Empty, Container, Text
                container = Container([Text("A"), Empty(), Text("B")])
        """
        self._style: str = style

        super().__init__(widget_id, file_path=__file__)

    def get_json_data(self) -> None:
        """Empty widget does not have any data and this method always returns None."""
        return None

    def get_json_state(self) -> None:
        """Empty widget does not have any state and this method always returns None."""
        return None



from typing import List, Optional

from supervisely import Annotation
from supervisely.app.widgets import GridGallery


class CompareAnnotations(GridGallery):
    """Side-by-side comparison of annotations for one image.
    Read about it in `Developer Portal <https://developer.supervisely.com/app-development/widgets/compare-data/compareannotations>`_.
    """

    def __init__(
        self,
        columns_number: int,
        default_opacity: Optional[float] = 0.5,
        fill_rectangle: Optional[bool] = True,
        border_width: Optional[int] = 3,
        view_height: Optional[int] = None,
        empty_message: Optional[str] = "No image was provided",
        widget_id: Optional[str] = None,
    ):
        """
        :param columns_number: Number of columns in the grid.
        :type columns_number: int
        :param default_opacity: Default opacity for annotations.
        :type default_opacity: float, optional
        :param fill_rectangle: if True, rectangles will be filled with color
        :type fill_rectangle: bool, optional
        :param border_width: Border width for rectangles.
        :type border_width: int, optional
        :param view_height: height of the widget
        :type view_height: int
        :param empty_message: message to show when there is no data
        :type empty_message: str
        :param widget_id: An unique identifier of the widget.
        :type widget_id: str, optional

        :Usage Example:

            .. code-block:: python

                from supervisely.app.widgets import CompareAnnotations

                compare_annotations = CompareAnnotations(columns_number=2)
                compare_annotations.set_image_url("https://i.imgur.com/2Yj2xYh.jpg")
                compare_annotations.append(ann1, "Annotation 1")
                compare_annotations.append(ann2, "Annotation 2")
        """
        self._image_url: str = None
        self._annotations: List[Annotation] = []
        self._titles: List[str] = []
        self._set_items(self._annotations)

        super().__init__(
            columns_number=columns_number,
            annotations_opacity=default_opacity,
            show_opacity_slider=True,
            enable_zoom=True,
            sync_views=True,
            fill_rectangle=fill_rectangle,
            border_width=border_width,
            show_preview=False,
            view_height=view_height,
            empty_message=empty_message,
            widget_id=widget_id,
        )

    @property
    def image_url(self) -> str:
        """Returns url of the image to show.

        :returns: url of the image to show
        :rtype: str
        """
        return self._image_url

    def set_image_url(self, image_url: str) -> None:
        """Sets url of the image to show.

        :param image_url: url of the image to show
        :type image_url: str
        """
        self._image_url = image_url

    def append(
        self,
        annotation: Optional[Annotation] = None,
        title: Optional[str] = "",
        column_index: Optional[int] = None,
    ) -> None:
        """Adds annotation to the widget.

        :param annotation: annotation to add
        :type annotation: :class:`~supervisely.annotation.annotation.Annotation`, optional
        :param title: title for the annotation
        :type title: str, optional
        :param column_index: index of the column to add annotation to
        :type column_index: int, optional
        """
        super().append(self._image_url, annotation, title, column_index)

    def is_empty(self) -> bool:
        """Returns True if there is no image to show, False otherwise.

        :returns: True if there is no image to show, False otherwise
        :rtype: bool
        """
        return self._image_url is None or self._image_url == ""

    def clean_up(self) -> None:
        """Removes all annotations and image url from the widget."""
        super().clean_up()
        self._image_url = None
        self._annotations = []
        self._titles = []

    def _set_items(self, annotations: List[Annotation]) -> None:
        for ann in annotations:
            self.append(self._image_url, ann)



import argparse
import json
import os
import shutil
import subprocess
import sys
import tempfile
import unittest
from datetime import datetime


class SetupTests(unittest.TestCase):
    """Test installation of supervisely package with different extras_require options.

    SAFETY GUARANTEES:
    - All tests run in ISOLATED temporary virtual environments
    - System Python and its packages are NEVER modified
    - All virtual environments are created in /tmp with unique names
    - Virtual environments are automatically cleaned up after tests
    - Package installation uses -e (editable) mode for local development
    """

    # List of all extras_require options from setup.py
    EXTRAS_OPTIONS = [
        "extras",
        "apps",
        "docs",
        "sdk-no-usages",
        "tracking",
        "model-benchmark",
        "training",
        "plugins",
        "sdk-nn-plugins",
        "aug",
        "agent",
    ]

    # Python versions to test (can be overridden by command line argument)
    PYTHON_VERSIONS = ["3.8", "3.9", "3.10", "3.11", "3.12", "3.13", "3.14"]

    # Filtered versions for testing (set by command line arguments)
    TEST_VERSIONS = None

    # Results storage
    test_results = {
        "start_time": None,
        "end_time": None,
        "python_versions": {},
        "tests": [],
        "summary": {"total": 0, "passed": 0, "failed": 0, "errors": 0},
    }

    @classmethod
    def _save_results(cls):
        """Save test results to a JSON file."""
        cls.test_results["end_time"] = datetime.now().isoformat()

        # Create results directory if it doesn't exist
        results_dir = os.path.join(cls.project_root, "test_results")
        os.makedirs(results_dir, exist_ok=True)

        # Generate filename with timestamp
        timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
        results_file = os.path.join(results_dir, f"setup_test_results_{timestamp}.json")

        # Save JSON results
        with open(results_file, "w") as f:
            json.dump(cls.test_results, f, indent=2)

        # Also create a human-readable report
        report_file = os.path.join(results_dir, f"setup_test_report_{timestamp}.txt")
        with open(report_file, "w") as f:
            f.write("=" * 80 + "\n")
            f.write("SUPERVISELY SDK INSTALLATION TEST REPORT\n")
            f.write("=" * 80 + "\n\n")
            f.write(f"Test Date: {cls.test_results['start_time']}\n")
            f.write(f"Test Duration: {cls.test_results['end_time']}\n\n")

            f.write("Python Versions Tested:\n")
            f.write("-" * 80 + "\n")
            for version, info in cls.test_results["python_versions"].items():
                f.write(f"  Python {version}: {info}\n")
            f.write("\n")

            f.write("Test Summary:\n")
            f.write("-" * 80 + "\n")
            f.write(f"  Total Tests: {cls.test_results['summary']['total']}\n")
            f.write(f"  Passed: {cls.test_results['summary']['passed']}\n")
            f.write(f"  Failed: {cls.test_results['summary']['failed']}\n")
            f.write(f"  Errors: {cls.test_results['summary']['errors']}\n\n")

            f.write("Detailed Results:\n")
            f.write("-" * 80 + "\n")
            for test in cls.test_results["tests"]:
                status_symbol = "✓" if test["status"] == "PASS" else "✗"
                f.write(f"{status_symbol} {test['name']} (Python {test['python_version']})\n")
                f.write(f"   Status: {test['status']}\n")
                if test.get("error"):
                    f.write(f"   Error: {test['error']}\n")
                f.write("\n")

        print(f"\n{'='*80}")
        print(f"Test results saved to:")
        print(f"  JSON: {results_file}")
        print(f"  Report: {report_file}")
        print(f"{'='*80}\n")

        return results_file, report_file

    @classmethod
    def _check_pyenv(cls):
        """Check if pyenv is available."""
        try:
            result = subprocess.run(["pyenv", "--version"], capture_output=True, text=True)
            return result.returncode == 0
        except FileNotFoundError:
            return False

    @classmethod
    def _version_key(cls, version_string):
        """Convert version string to tuple of integers for proper sorting.

        Examples:
            '3.8.9' -> (3, 8, 9)
            '3.8.20' -> (3, 8, 20)
            '3.10.1' -> (3, 10, 1)
        """
        # Remove any leading/trailing whitespace
        version_string = version_string.strip()

        # Split by dots and convert to integers
        parts = version_string.split(".")
        try:
            return tuple(int(p) for p in parts if p.isdigit())
        except ValueError:
            # If conversion fails, return original string (will be sorted lexicographically)
            return (0,)  # Put invalid versions at the beginning

    @classmethod
    def _get_pyenv_versions(cls):
        """Get list of installed Python versions from pyenv."""
        try:
            result = subprocess.run(["pyenv", "versions", "--bare"], capture_output=True, text=True)
            if result.returncode == 0:
                return [v.strip() for v in result.stdout.strip().split("\n") if v.strip()]
            return []
        except FileNotFoundError:
            return []

    @classmethod
    def _install_python_with_pyenv(cls, version):
        """Install Python version using pyenv."""
        print(f"\nInstalling Python {version} using pyenv...")
        print(f"This may take several minutes...\n")

        result = subprocess.run(
            ["pyenv", "install", "-s", version], stdout=sys.stdout, stderr=sys.stderr
        )

        return result.returncode == 0

    @classmethod
    def _find_pyenv_python_version(cls, target_version):
        """Find matching Python version in pyenv for target version (e.g., '3.8' -> '3.8.18').
        Returns only INSTALLED versions, not available for installation."""
        installed_versions = cls._get_pyenv_versions()

        # Find all versions matching the target, excluding virtual environments
        matching = [
            v
            for v in installed_versions
            if v.startswith(target_version + ".")
            and "/" not in v  # Exclude virtual environments like "3.9.22/envs/.venv"
            and not any(x in v for x in ["-", "a", "b", "rc", "dev"])  # Exclude dev versions
        ]

        if matching:
            # Return the latest patch version using proper version comparison
            return sorted(matching, key=cls._version_key)[-1]

        # Version not installed
        return None

    @classmethod
    def _find_installable_python_version(cls, target_version):
        """Find the latest installable Python version for target version.
        Returns version string that can be installed via pyenv."""
        try:
            result = subprocess.run(["pyenv", "install", "--list"], capture_output=True, text=True)
            if result.returncode == 0:
                available = [
                    v.strip()
                    for v in result.stdout.strip().split("\n")
                    if v.strip().startswith(target_version + ".")
                    and not any(x in v for x in ["-", "a", "b", "rc", "dev", "/"])
                ]
                if available:
                    latest = sorted(available, key=cls._version_key)[-1]
                    return latest
        except Exception as e:
            print(f"Error finding available versions: {e}")

        return None

    @classmethod
    def _get_project_version(cls):
        """Get project version from git branch and latest tag."""
        try:
            # Get current branch
            branch = subprocess.run(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                capture_output=True,
                text=True,
                cwd=cls.project_root,
            ).stdout.strip()

            # Get latest tag
            tag = subprocess.run(
                ["git", "describe", "--tags", "--abbrev=0"],
                capture_output=True,
                text=True,
                cwd=cls.project_root,
            ).stdout.strip()

            if branch and branch != "master" and branch != "main":
                return f"{tag}+{branch}" if tag else f"0.0.0+{branch}"
            return tag if tag else "0.0.0"
        except Exception as e:
            print(f"Warning: Could not determine version from git: {e}")
            return "0.0.0+test"

    @classmethod
    def setUpClass(cls):
        """Create test virtual environments for all Python versions using pyenv."""
        cls.test_results["start_time"] = datetime.now().isoformat()
        cls.test_venv_dirs = {}
        cls.test_pythons = {}
        cls.test_pips = {}
        cls.project_root = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

        # Get project version to avoid GitHub API calls
        cls.release_version = cls._get_project_version()
        print(f"\nUsing version: {cls.release_version}\n")

        # Determine which versions to test
        versions_to_test = cls.TEST_VERSIONS if cls.TEST_VERSIONS else cls.PYTHON_VERSIONS
        print(f"Requested Python versions: {', '.join(versions_to_test)}")

        print(f"\n{'='*80}")
        print(f"Setting up test environments for Python versions: {', '.join(versions_to_test)}")
        print(f"{'='*80}\n")

        # Check if pyenv is available
        if not cls._check_pyenv():
            print("⚠ pyenv is not installed or not in PATH")
            print("Falling back to system Python versions...\n")
            use_pyenv = False
        else:
            pyenv_version = subprocess.run(["pyenv", "--version"], capture_output=True, text=True)
            print(f"✓ Found {pyenv_version.stdout.strip()}\n")
            use_pyenv = True

        # Find available Python versions
        cls.available_pythons = {}

        if use_pyenv:
            # Use pyenv to find or install Python versions
            for version in versions_to_test:
                print(f"Looking for Python {version} in pyenv...")

                # Check if version is already installed
                full_version = cls._find_pyenv_python_version(version)

                if full_version:
                    print(f"✓ Found Python {full_version} in pyenv")
                    cls.available_pythons[version] = full_version
                    cls.test_results["python_versions"][version] = f"pyenv {full_version}"
                else:
                    # Version not installed, try to find and install it
                    print(f"Python {version} not installed, searching for available versions...")

                    installable_version = cls._find_installable_python_version(version)

                    if installable_version:
                        print(f"Found installable version: {installable_version}")
                        print(
                            f"Installing Python {installable_version} with pyenv (this may take several minutes)..."
                        )

                        install_result = subprocess.run(
                            ["pyenv", "install", "-s", installable_version],
                            stdout=sys.stdout,
                            stderr=sys.stderr,
                        )

                        if install_result.returncode == 0:
                            print(f"✓ Successfully installed Python {installable_version}")
                            cls.available_pythons[version] = installable_version
                            cls.test_results["python_versions"][
                                version
                            ] = f"pyenv {installable_version} (installed)"
                        else:
                            print(f"✗ Failed to install Python {installable_version}")
                    else:
                        print(f"✗ No installable version found for Python {version}")
        else:
            # Fallback to system Python
            for version in versions_to_test:
                python_cmd = f"python{version}"
                try:
                    result = subprocess.run(
                        [python_cmd, "--version"], capture_output=True, text=True
                    )
                    if result.returncode == 0:
                        cls.available_pythons[version] = python_cmd
                        cls.test_results["python_versions"][
                            version
                        ] = f"system {result.stdout.strip()}"
                        print(f"✓ Found {result.stdout.strip()}")
                    else:
                        print(f"✗ Python {version} not found")
                except FileNotFoundError:
                    print(f"✗ Python {version} not found")

        if not cls.available_pythons:
            raise RuntimeError(
                "No Python versions found! Please install Python 3.8-3.14 using pyenv or system package manager"
            )

        print(f"\n{'='*80}")
        print(f"Creating virtual environments...")
        print(f"{'='*80}\n")

        # Create virtual environments for each available Python version
        for version, python_identifier in cls.available_pythons.items():
            venv_dir = tempfile.mkdtemp(prefix=f"test_venv_py{version}_")

            # SAFETY CHECK: Ensure we're creating venv in /tmp
            if not venv_dir.startswith("/tmp/") and not venv_dir.startswith(tempfile.gettempdir()):
                raise RuntimeError(
                    f"SAFETY ERROR: Virtual environment must be in /tmp, got: {venv_dir}"
                )

            print(f"Creating venv for Python {version} at: {venv_dir}")

            try:
                if use_pyenv:
                    # Use pyenv to get the Python executable path
                    result_prefix = subprocess.run(
                        ["pyenv", "prefix", python_identifier], capture_output=True, text=True
                    )

                    if result_prefix.returncode != 0:
                        print(f"✗ Failed to get pyenv prefix for {python_identifier}")
                        print(f"Error: {result_prefix.stderr}")
                        shutil.rmtree(venv_dir, ignore_errors=True)
                        continue

                    python_path = result_prefix.stdout.strip()

                    # Ensure we have an absolute path
                    if not os.path.isabs(python_path):
                        python_path = os.path.abspath(python_path)

                    python_bin = os.path.join(python_path, "bin", "python")

                    if not os.path.exists(python_bin):
                        print(f"✗ Python executable not found at: {python_bin}")
                        shutil.rmtree(venv_dir, ignore_errors=True)
                        continue

                    print(f"Using Python: {python_bin}")

                    result = subprocess.run(
                        [python_bin, "-m", "venv", venv_dir], stdout=sys.stdout, stderr=sys.stderr
                    )
                else:
                    result = subprocess.run(
                        [python_identifier, "-m", "venv", venv_dir],
                        stdout=sys.stdout,
                        stderr=sys.stderr,
                    )

                if result.returncode != 0:
                    print(f"✗ Failed to create virtual environment for Python {version}")
                    shutil.rmtree(venv_dir, ignore_errors=True)
                    continue

                # Successfully created venv, add to tracking
                cls.test_venv_dirs[version] = venv_dir

            except Exception as e:
                print(f"✗ Exception creating venv for Python {version}: {e}")
                shutil.rmtree(venv_dir, ignore_errors=True)
                continue

            # Get paths for the test virtual environment
            if sys.platform.startswith("win"):
                cls.test_pythons[version] = os.path.join(venv_dir, "Scripts", "python.exe")
                cls.test_pips[version] = os.path.join(venv_dir, "Scripts", "pip.exe")
            else:
                cls.test_pythons[version] = os.path.join(venv_dir, "bin", "python")
                cls.test_pips[version] = os.path.join(venv_dir, "bin", "pip")

            # Upgrade pip
            print(f"Upgrading pip for Python {version}...")
            subprocess.run(
                [cls.test_pips[version], "install", "--upgrade", "pip", "-q"],
                stdout=sys.stdout,
                stderr=sys.stderr,
            )
            print(f"✓ Environment ready for Python {version}\n")

    @classmethod
    def tearDownClass(cls):
        """Remove all test virtual environments."""
        print(f"\n{'='*80}")
        print(f"TEARDOWN: Cleaning up test virtual environments")
        print(f"Total tests run: {cls.test_results['summary']['total']}")
        print(f"{'='*80}\n")
        for version, venv_dir in cls.test_venv_dirs.items():
            print(f"Removing test environment for Python {version}: {venv_dir}")
            if os.path.exists(venv_dir):
                shutil.rmtree(venv_dir)
            else:
                print(f"  WARNING: venv already removed: {venv_dir}")

        # Save test results
        cls._save_results()

    def _test_install_with_extra(self, extra_name, python_version):
        """Helper method to test installation with a specific extra for a specific Python version."""
        test_name = f"test_install_{extra_name}_py{python_version}"
        test_result = {
            "name": extra_name,
            "python_version": python_version,
            "status": "PASS",
            "error": None,
        }

        try:
            print(f"\n{'='*80}")
            print(f"Testing installation with extra: [{extra_name}] for Python {python_version}")
            print(f"{'='*80}\n")

            # Check if Python version is available
            if python_version not in self.test_pips or python_version not in self.test_pythons:
                raise RuntimeError(
                    f"Python {python_version} environment not available. "
                    f"Available versions: {list(self.test_pips.keys())}"
                )

            test_pip = self.test_pips[python_version]
            test_python = self.test_pythons[python_version]

            # SAFETY CHECK: Ensure venv still exists
            venv_dir = self.test_venv_dirs[python_version]
            if not os.path.exists(venv_dir):
                raise RuntimeError(
                    f"Virtual environment for Python {python_version} does not exist at: {venv_dir}"
                )

            # SAFETY CHECK: Ensure pip and python executables still exist
            if not os.path.exists(test_pip):
                raise RuntimeError(f"pip executable does not exist at: {test_pip}")
            if not os.path.exists(test_python):
                raise RuntimeError(f"python executable does not exist at: {test_python}")

            print(f"✓ Environment validated:")
            print(f"  venv: {venv_dir}")
            print(f"  python: {test_python}")
            print(f"  pip: {test_pip}\n")

            # SAFETY CHECK: Ensure we're using pip from isolated venv
            if not test_pip.startswith(venv_dir):
                raise RuntimeError(
                    f"SAFETY ERROR: pip must be from isolated venv!\n"
                    f"  Expected prefix: {venv_dir}\n"
                    f"  Got: {test_pip}"
                )

            if not test_python.startswith(venv_dir):
                raise RuntimeError(
                    f"SAFETY ERROR: python must be from isolated venv!\n"
                    f"  Expected prefix: {venv_dir}\n"
                    f"  Got: {test_python}"
                )

            # Install the package with the specific extra
            install_cmd = [
                test_pip,
                "install",
                "-e",
                f"{self.project_root}[{extra_name}]",
            ]

            print(f"Running: {' '.join(install_cmd)}\n")

            # Set environment variable to avoid GitHub API calls in setup.py
            env = os.environ.copy()
            env["RELEASE_VERSION"] = self.release_version

            install_result = subprocess.run(
                install_cmd, stdout=sys.stdout, stderr=sys.stderr, text=True, env=env
            )

            # Check if the installation was successful
            self.assertEqual(
                install_result.returncode,
                0,
                msg=f"Installation with [{extra_name}] for Python {python_version} failed with return code {install_result.returncode}",
            )

            # Verify supervisely can be imported
            print(f"\n{'='*80}")
            print(f"Verifying import for [{extra_name}] (Python {python_version})")
            print(f"{'='*80}\n")

            import_cmd = [
                test_python,
                "-c",
                "import supervisely; print(f'Successfully imported supervisely {supervisely.__version__}')",
            ]
            import_result = subprocess.run(
                import_cmd, stdout=sys.stdout, stderr=sys.stderr, text=True
            )

            self.assertEqual(
                import_result.returncode,
                0,
                msg=f"Failed to import supervisely with [{extra_name}] (Python {python_version})",
            )

            # List all installed packages and their versions
            print(f"\n{'='*80}")
            print(f"Installed packages for [{extra_name}] (Python {python_version})")
            print(f"{'='*80}\n")

            list_packages_cmd = [test_pip, "list"]
            subprocess.run(list_packages_cmd, stdout=sys.stdout, stderr=sys.stderr)

            print(f"\n{'='*80}")
            print(
                f"✓ Successfully tested installation with [{extra_name}] for Python {python_version}"
            )
            print(f"{'='*80}\n")

            test_result["status"] = "PASS"
            self.test_results["summary"]["passed"] += 1

        except Exception as e:
            test_result["status"] = "FAIL"
            test_result["error"] = str(e)
            self.test_results["summary"]["failed"] += 1
            raise
        finally:
            self.test_results["tests"].append(test_result)
            self.test_results["summary"]["total"] += 1

    def test_01_base_installation(self):
        """Test base installation without any extras for all Python versions."""
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                test_result = {
                    "name": "base_installation",
                    "python_version": version,
                    "status": "PASS",
                    "error": None,
                }

                try:
                    print(f"\n{'='*80}")
                    print(f"Testing base installation (no extras) for Python {version}")
                    print(f"{'='*80}\n")

                    # Check if Python version is available
                    if version not in self.test_pips or version not in self.test_pythons:
                        raise RuntimeError(
                            f"Python {version} environment not available. "
                            f"Available versions: {list(self.test_pips.keys())}"
                        )

                    test_pip = self.test_pips[version]
                    test_python = self.test_pythons[version]

                    # SAFETY CHECK: Ensure venv still exists
                    venv_dir = self.test_venv_dirs[version]
                    if not os.path.exists(venv_dir):
                        raise RuntimeError(
                            f"Virtual environment for Python {version} does not exist at: {venv_dir}"
                        )

                    # SAFETY CHECK: Ensure pip and python executables still exist
                    if not os.path.exists(test_pip):
                        raise RuntimeError(f"pip executable does not exist at: {test_pip}")
                    if not os.path.exists(test_python):
                        raise RuntimeError(f"python executable does not exist at: {test_python}")

                    print(f"✓ Environment validated:")
                    print(f"  venv: {venv_dir}")
                    print(f"  python: {test_python}")
                    print(f"  pip: {test_pip}\n")

                    # SAFETY CHECK: Ensure we're using pip/python from isolated venv
                    if not test_pip.startswith(venv_dir):
                        raise RuntimeError(
                            f"SAFETY ERROR: pip must be from isolated venv!\n"
                            f"  Expected prefix: {venv_dir}\n"
                            f"  Got: {test_pip}"
                        )

                    if not test_python.startswith(venv_dir):
                        raise RuntimeError(
                            f"SAFETY ERROR: python must be from isolated venv!\n"
                            f"  Expected prefix: {venv_dir}\n"
                            f"  Got: {test_python}"
                        )

                    install_cmd = [
                        test_pip,
                        "install",
                        "--no-cache-dir",
                        "-e",
                        self.project_root,
                    ]

                    print(f"Running: {' '.join(install_cmd)}\n")

                    # Set environment variable to avoid GitHub API calls in setup.py
                    env = os.environ.copy()
                    env["RELEASE_VERSION"] = self.release_version

                    install_result = subprocess.run(
                        install_cmd, stdout=sys.stdout, stderr=sys.stderr, text=True, env=env
                    )

                    self.assertEqual(
                        install_result.returncode,
                        0,
                        msg=f"Base installation failed for Python {version}",
                    )

                    # Check if supervisely is installed
                    print(f"\n{'='*80}")
                    print(f"Checking installed packages for Python {version}")
                    print(f"{'='*80}\n")

                    list_cmd = [test_pip, "list"]
                    list_result = subprocess.run(list_cmd, capture_output=True, text=True)

                    self.assertIn(
                        "supervisely",
                        list_result.stdout,
                        msg=f"supervisely is not in the list of installed packages for Python {version}",
                    )
                    print(f"✓ supervisely found in installed packages for Python {version}")

                    # Check if the supervisely package can be imported
                    print(f"\n{'='*80}")
                    print(f"Verifying import for Python {version}")
                    print(f"{'='*80}\n")

                    import_cmd = [
                        test_python,
                        "-c",
                        "import supervisely; print(f'Successfully imported supervisely {supervisely.__version__}')",
                    ]
                    import_result = subprocess.run(
                        import_cmd, stdout=sys.stdout, stderr=sys.stderr, text=True
                    )

                    self.assertEqual(
                        import_result.returncode,
                        0,
                        msg=f"Failed to import supervisely package for Python {version}",
                    )

                    # List all installed packages and their versions
                    print(f"\n{'='*80}")
                    print(f"Installed packages for Python {version}")
                    print(f"{'='*80}\n")

                    list_packages_cmd = [test_pip, "list"]
                    subprocess.run(list_packages_cmd, stdout=sys.stdout, stderr=sys.stderr)

                    print(f"\n{'='*80}")
                    print(f"✓ Base installation test passed for Python {version}")
                    print(f"{'='*80}\n")

                    test_result["status"] = "PASS"
                    self.test_results["summary"]["passed"] += 1

                except Exception as e:
                    test_result["status"] = "FAIL"
                    test_result["error"] = str(e)
                    self.test_results["summary"]["failed"] += 1
                    raise
                finally:
                    self.test_results["tests"].append(test_result)
                    self.test_results["summary"]["total"] += 1

    def test_02_extras(self):
        """Test installation with 'extras' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 02: EXTRAS")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("extras", version)

    def test_03_apps(self):
        """Test installation with 'apps' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 03: APPS")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("apps", version)

    def test_04_docs(self):
        """Test installation with 'docs' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 04: DOCS")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("docs", version)

    def test_05_sdk_no_usages(self):
        """Test installation with 'sdk-no-usages' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 05: SDK-NO-USAGES")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("sdk-no-usages", version)

    def test_06_tracking(self):
        """Test installation with 'tracking' option for all Python versions."""
        if sys.platform == "darwin":
            self.skipTest("Skipping [tracking] extra on macOS (faiss-gpu is not available)")
        print(f"\n{'#'*80}")
        print(f"# TEST 06: TRACKING")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("tracking", version)

    def test_07_model_benchmark(self):
        """Test installation with 'model-benchmark' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 07: MODEL-BENCHMARK")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("model-benchmark", version)

    def test_08_training(self):
        """Test installation with 'training' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 08: TRAINING")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("training", version)

    def test_09_plugins(self):
        """Test installation with 'plugins' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 09: PLUGINS")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("plugins", version)

    def test_10_sdk_nn_plugins(self):
        """Test installation with 'sdk-nn-plugins' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 10: SDK-NN-PLUGINS")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("sdk-nn-plugins", version)

    def test_11_aug(self):
        """Test installation with 'aug' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 11: AUG")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("aug", version)

    def test_12_agent(self):
        """Test installation with 'agent' option for all Python versions."""
        print(f"\n{'#'*80}")
        print(f"# TEST 12: AGENT")
        print(f"{'#'*80}\n")
        for version in self.available_pythons.keys():
            with self.subTest(python_version=version):
                self._test_install_with_extra("agent", version)


if __name__ == "__main__":
    # Parse command line arguments
    parser = argparse.ArgumentParser(
        description="Test Supervisely SDK installation with different extras",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  # Run all tests for all Python versions
  python3 tests/test_setup.py
  
  # Run tests only for Python 3.11 and 3.12
  python3 tests/test_setup.py --python-versions 3.11 3.12
  
  # Run specific test for specific Python versions
  python3 tests/test_setup.py --python-versions 3.11 3.12 SetupTests.test_06_tracking
  
  # Run with verbose output
  python3 tests/test_setup.py --python-versions 3.11 -v
        """,
    )
    parser.add_argument(
        "--python-versions",
        nargs="+",
        metavar="VERSION",
        help="Python versions to test (e.g., 3.8 3.9 3.10 3.11 3.12 3.13 3.14)",
    )

    # Parse known args to allow unittest args to pass through
    args, remaining = parser.parse_known_args()

    # Set the test versions if specified
    if args.python_versions:
        # Separate Python versions from test names
        # Python versions are in format X.Y or X.YY
        python_versions = []
        test_names = []

        for arg in args.python_versions:
            # Check if it looks like a Python version (e.g., "3.8", "3.11")
            if arg.replace(".", "").isdigit() and arg.count(".") == 1:
                python_versions.append(arg)
            else:
                # Treat as test name
                test_names.append(arg)

        # Validate Python versions
        if python_versions:
            invalid_versions = [v for v in python_versions if v not in SetupTests.PYTHON_VERSIONS]
            if invalid_versions:
                print(f"Error: Invalid Python version(s): {', '.join(invalid_versions)}")
                print(f"Valid versions are: {', '.join(SetupTests.PYTHON_VERSIONS)}")
                sys.exit(1)

            SetupTests.TEST_VERSIONS = python_versions
            print(f"Testing only Python versions: {', '.join(python_versions)}\n")

        # Add test names to remaining args for unittest
        if test_names:
            remaining = test_names + remaining
    else:
        print(f"Testing all Python versions: {', '.join(SetupTests.PYTHON_VERSIONS)}\n")

    # Run unittest with remaining arguments
    sys.argv = [sys.argv[0]] + remaining
    unittest.main(verbosity=2)



from typing import Callable


class BaseTrainLogger:
    """Abstract base for training loggers with callbacks for train/epoch/step lifecycle events."""

    def __init__(self):
        self._on_train_started_callbacks = []
        self._on_train_finished_callbacks = []
        self._on_epoch_started_callbacks = []
        self._on_epoch_finished_callbacks = []
        self._on_step_finished_callbacks = []
        self.epoch_idx = 0
        self.step_idx = 0

    def train_started(self, total_epochs: int):
        for callback in self._on_train_started_callbacks:
            callback(total_epochs)

    def train_finished(self):
        for callback in self._on_train_finished_callbacks:
            callback()

    def epoch_started(self, total_steps: int):
        for callback in self._on_epoch_started_callbacks:
            callback(total_steps)

    def epoch_finished(self):
        self.epoch_idx += 1
        for callback in self._on_epoch_finished_callbacks:
            callback()

    def step_finished(self):
        self.step_idx += 1
        for callback in self._on_step_finished_callbacks:
            callback()

    def add_on_train_started_callback(self, callback: Callable):
        self._on_train_started_callbacks.append(callback)

    def add_on_train_finish_callback(self, callback: Callable):
        self._on_train_finished_callbacks.append(callback)

    def add_on_epoch_started_callback(self, callback: Callable):
        self._on_epoch_started_callbacks.append(callback)

    def add_on_epoch_finish_callback(self, callback: Callable):
        self._on_epoch_finished_callbacks.append(callback)

    def add_on_step_finished_callback(self, callback: Callable):
        self._on_step_finished_callbacks.append(callback)

    def log(self, logs: dict, idx: int):
        raise NotImplementedError

    def log_step(self, logs: dict):
        raise NotImplementedError

    def log_epoch(self, logs: dict):
        raise NotImplementedError
    
    def close(self):
        pass
